@@ -13,6 +13,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 import argparse
+import distutils.util
 import glob
 import itertools
 import logging
@@ -96,10 +97,31 @@ def git_remote_matching_url(url):
             return remote.partition("\t")[0]
 
 
-def git_remote_url(remote="origin"):
+def git_remote_url(remote="origin", raise_on_error=True):
     return _run_shell_command(
         ["git", "config", "--get", "remote." + remote + ".url"],
-        output=True)
+        output=True, raise_on_error=raise_on_error)
+
+
+def git_get_config(option, default):
+    try:
+        return _run_shell_command(
+            ["git", "config", "--get", "git-pull-request." + option],
+            output=True)
+    except RuntimeError:
+        return default
+
+
+def git_config_add_argument(parser, option, *args, **kwargs):
+    default = kwargs.get("default")
+    isboolean = kwargs.get("action") in ["store_true", "store_false"]
+    if isboolean and default is None:
+        default = False
+    default = git_get_config(option[2:], default)
+    if isboolean and isinstance(default, str):
+        default = distutils.util.strtobool(default)
+    kwargs["default"] = default
+    return parser.add_argument(option, *args, **kwargs)
 
 
 def git_get_branch_name():
@@ -187,7 +209,7 @@ def parse_pr_message(message):
 
 def git_get_commit_body(commit):
     return _run_shell_command(
-        ["git", "show", "-q", "--format=%b", commit],
+        ["git", "show", "-q", "--format=%b", commit, '--'],
         output=True)
 
 
@@ -227,8 +249,12 @@ def git_pull_request(target_remote=None, target_branch=None,
                      rebase=True,
                      force_editor=False,
                      download=None,
+                     download_setup=False,
                      tag_previous_revision=False,
-                     dont_fork=True):
+                     fork=True,
+                     setup_only=False,
+                     branch_prefix=None,
+                     dry_run=False):
     branch = git_get_branch_name()
     if not branch:
         LOG.critical("Unable to find current branch")
@@ -293,20 +319,29 @@ def git_pull_request(target_remote=None, target_branch=None,
         repo = g.get_user(user_to_fork).get_repo(reponame_to_fork)
 
     if download is not None:
-        download_pull_request(g, repo, target_remote, download)
+        retcode = download_pull_request(g, repo, target_remote,
+                                        download, download_setup)
+
     else:
-        fork_and_push_pull_request(g, hosttype, repo, rebase, target_remote,
-                                   target_branch, branch, user, title, message,
-                                   comment,
-                                   force_editor, tag_previous_revision,
-                                   dont_fork)
+        retcode = fork_and_push_pull_request(
+            g, hosttype, repo, rebase, target_remote, target_branch, branch,
+            user, title, message, comment, force_editor, tag_previous_revision,
+            fork, setup_only, branch_prefix, dry_run,
+        )
+
     approve_login_password(host=hostname, user=user, password=password)
 
+    return retcode
 
-def download_pull_request(g, repo, target_remote, pull_number):
+
+def download_pull_request(g, repo, target_remote, pull_number,
+                          setup_remote):
     pull = repo.get_pull(pull_number)
-    local_branch_name = "pull/%d-%s-%s" % (pull.number, pull.user.login,
-                                           pull.head.ref)
+    if setup_remote:
+        local_branch_name = pull.head.ref
+    else:
+        local_branch_name = "pull/%d-%s-%s" % (pull.number, pull.user.login,
+                                               pull.head.ref)
     target_ref = "pull/%d/head" % pull.number
 
     _run_shell_command(["git", "fetch", target_remote, target_ref])
@@ -318,6 +353,15 @@ def download_pull_request(g, repo, target_remote, pull_number):
     else:
         _run_shell_command(["git", "reset", "--hard", "FETCH_HEAD"])
 
+    if setup_remote:
+        remote_name = "github-%s" % pull.user.login
+        remote = git_remote_url(remote_name, raise_on_error=False)
+        if not remote:
+            _run_shell_command(["git", "remote", "add", remote_name,
+                                pull.head.repo.clone_url])
+        _run_shell_command(["git", "fetch", remote_name])
+        _run_shell_command(["git", "branch", "-u", "origin", pull.base.ref])
+
 
 def edit_file_get_content_and_remove(filename):
     editor = _run_shell_command(["git", "var", "GIT_EDITOR"], output=True)
@@ -385,17 +429,39 @@ def preserve_older_revision(branch, remote_to_push):
 def fork_and_push_pull_request(g, hosttype, repo_to_fork, rebase,
                                target_remote, target_branch, branch, user,
                                title, message, comment,
-                               force_editor, tag_previous_revision, dont_fork):
+                               force_editor, tag_previous_revision, fork,
+                               setup_only, branch_prefix,
+                               dry_run=False):
 
     g_user = g.get_user()
 
-    if dont_fork:
-        remote_to_push = target_remote
-        head = "{}:{}".format(repo_to_fork.owner.login, branch)
-    else:
-        repo_forked = g_user.create_fork(repo_to_fork)
-        LOG.info("Forked repository: %s", repo_forked.html_url)
+    forked = False
+    if fork in ["always", "auto"]:
+        try:
+            repo_forked = g_user.create_fork(repo_to_fork)
+        except github.GithubException as e:
+            if (fork == "auto" and e.status == 403 and
+                    "forking is disabled" in e.data["message"]):
+                forked = False
+                LOG.info("Forking is disabled on target repository, "
+                         "using base repository")
+            else:
+                LOG.error("Forking is disabled on target repository, "
+                          "can't fork", exc_info=True)
+                sys.exit(1)
+        else:
+            forked = True
+            LOG.info("Forked repository: %s", repo_forked.html_url)
 
+    if branch_prefix is None and not forked:
+        branch_prefix = g_user.login
+
+    if branch_prefix:
+        remote_branch = "{}/{}".format(branch_prefix, branch)
+    else:
+        remote_branch = branch
+
+    if forked:
         remote_to_push = git_remote_matching_url(repo_forked.clone_url)
 
         if remote_to_push:
@@ -408,6 +474,14 @@ def fork_and_push_pull_request(g, hosttype, repo_to_fork, rebase,
                  remote_to_push, repo_forked.clone_url])
             LOG.info("Added forked repository as remote `%s'", remote_to_push)
         head = "{}:{}".format(user, branch)
+    else:
+        remote_to_push = target_remote
+        head = "{}:{}".format(repo_to_fork.owner.login, remote_branch)
+
+    if setup_only:
+        LOG.info("Fetch existing branches of remote `%s`", remote_to_push)
+        _run_shell_command(["git", "fetch", remote_to_push])
+        return
 
     if rebase:
         _run_shell_command(["git", "remote", "update", target_remote])
@@ -433,13 +507,21 @@ def fork_and_push_pull_request(g, hosttype, repo_to_fork, rebase,
                 "resolved.")
             return 37
 
-    LOG.info("Force-pushing branch `%s' to remote `%s'",
-             branch, remote_to_push)
-
-    _run_shell_command(["git", "push", "-f", remote_to_push, branch])
+    if dry_run:
+        LOG.info("Would force-push branch `%s' to remote `%s/%s'",
+                 branch, remote_to_push, remote_branch)
+        return
+    else:
+        LOG.info("Force-pushing branch `%s' to remote `%s/%s'",
+                 branch, remote_to_push, remote_branch)
+        _run_shell_command([
+            "git", "push", "-f",
+            remote_to_push,
+            "{}:{}".format(branch, remote_branch),
+        ])
 
     pulls = list(repo_to_fork.get_pulls(base=target_branch,
-                                        head=user + ":" + branch))
+                                        head=head))
     if pulls:
         for pull in pulls:
             LOG.info("Pull-request updated:\n  %s", pull.html_url)
@@ -510,30 +592,53 @@ def _format_github_exception(action, exc):
     )
 
 
-def main():
+class DownloadAndSetupAction(argparse.Action):
+    def __call__(self, parser, namespace, values, option_strings=None):
+        setattr(namespace, "download", values)
+        if self.dest == "download":
+            setattr(namespace, "download_setup", False)
+        else:
+            setattr(namespace, "download_setup", True)
+
+
+def build_parser():
     parser = argparse.ArgumentParser(
         description='Send GitHub pull-request.'
     )
     parser.add_argument("--download", "-d",
                         type=int,
+                        action=DownloadAndSetupAction,
                         help="Checkout a pull request")
+    parser.add_argument("--download-and-setup", "-D",
+                        type=int,
+                        dest="download_setup",
+                        action=DownloadAndSetupAction,
+                        help=("Checkout a pull request and setup remote "
+                              "to be able to repush it"))
     parser.add_argument("--debug",
                         action='store_true',
                         help="Enabled debugging.")
-    parser.add_argument("--target-remote",
-                        help="Remote to send a pull-request to. "
-                        "Default is auto-detected from .git/config.")
-    parser.add_argument("--target-branch",
-                        help="Branch to send a pull-request to. "
-                        "Default is auto-detected from .git/config.")
+    parser.add_argument("--dry-run", "-n",
+                        action='store_true',
+                        help="Do not push not create the pull request.")
+    git_config_add_argument(parser, "--target-remote",
+                            help="Remote to send a pull-request to. "
+                            "Default is auto-detected from .git/config.")
+    git_config_add_argument(parser, "--target-branch",
+                            help="Branch to send a pull-request to. "
+                            "Default is auto-detected from .git/config.")
     parser.add_argument("--title",
                         help="Title of the pull request.")
     parser.add_argument("--message", "-m",
                         help="Message of the pull request.")
-    parser.add_argument("--no-rebase", "-R",
-                        action="store_true",
-                        help="Don't rebase branch before pushing.")
-    parser.add_argument(
+    git_config_add_argument(parser,
+                            "--branch-prefix",
+                            help="Prefix remote branch")
+    git_config_add_argument(parser, "--no-rebase", "-R",
+                            action="store_true",
+                            help="Don't rebase branch before pushing.")
+    git_config_add_argument(
+        parser,
         "--force-editor",
         action="store_true",
         default=False,
@@ -542,20 +647,44 @@ def main():
         "--comment", "-C",
         help="Comment to publish when updating the pull-request"
     )
-    parser.add_argument(
+    git_config_add_argument(
+        parser,
         "--tag-previous-revision",
         action="store_true",
         default=False,
         help="Preserve older revision when pushing"
     )
-    parser.add_argument(
+    group = parser.add_mutually_exclusive_group()
+    git_config_add_argument(
+        group,
+        "--fork",
+        default="auto",
+        choices=["always", "never", "auto"],
+        help=("Fork behavior to create the pull-request "
+              "(auto: when repository can't be cloned, "
+              "always: always try to fork it "
+              "never: always use base repository)")
+    )
+    git_config_add_argument(
+        group,
         "--no-fork",
-        action="store_true",
-        default=False,
+        dest="fork",
+        action="store_const",
+        const="never",
         help="Don't fork to create the pull-request"
     )
+    git_config_add_argument(
+        parser,
+        "--setup-only",
+        action="store_true",
+        default=False,
+        help="Just setup the fork repo"
+    )
+    return parser
 
-    args = parser.parse_args()
+
+def main():
+    args = build_parser().parse_args()
 
     daiquiri.setup(
         outputs=(
@@ -576,8 +705,12 @@ def main():
             rebase=not args.no_rebase,
             force_editor=args.force_editor,
             download=args.download,
+            download_setup=args.download_setup,
             tag_previous_revision=args.tag_previous_revision,
-            dont_fork=args.no_fork
+            fork=args.fork,
+            setup_only=args.setup_only,
+            branch_prefix=args.branch_prefix,
+            dry_run=args.dry_run,
         )
     except Exception:
         LOG.error("Unable to send pull request", exc_info=True)
